#================
# FILE          : KIWIURL.pm
#----------------
# PROJECT       : OpenSUSE Build-Service
# COPYRIGHT     : (c) 2006 SUSE LINUX Products GmbH, Germany
#               :
# AUTHOR        : Marcus Schaefer <ms@suse.de>
#               :
# BELONGS TO    : Operating System images
#               :
# DESCRIPTION   : This module is used to support the high level
#               : source locations like opensuse://
#               :
# STATUS        : Development
#----------------
package KIWIURL;
#==========================================
# Modules
#------------------------------------------
use strict;
use Carp qw (cluck);
use File::Basename;
use KIWILog;
use LWP;
use KIWIQX;

#==========================================
# Constructor
#------------------------------------------
sub new { 
	# ...
	# Create a new KIWIURL object which is used to solve
	# the high level location information into a low level
	# distribution independent network url
	# ---
	#==========================================
	# Object setup
	#------------------------------------------
	my $this  = {};
	my $class = shift;
	bless $this,$class;
	#==========================================
	# Module Parameters
	#------------------------------------------	
	my $kiwi = shift;
	my $root = shift;
	#==========================================
	# Constructor setup
	#------------------------------------------
	if (! defined $kiwi) {
		$kiwi = new KIWILog();
	}
	#==========================================
	# Store object data
	#------------------------------------------
	$this->{kiwi} = $kiwi;
	$this->{root} = $root;
	$this->{type} = "unknown";
	return $this;
}

#==========================================
# getRepoType
#------------------------------------------
sub getRepoType {
	# ...
	# return repo type, could be unknown
	# ---
	my $this = shift;
	return $this->{type};
}

#==========================================
# normalizePath
#------------------------------------------
sub normalizePath {
	# ...
	# check all path functions and normalize the high level
	# URLs if required. This function also quotes reserved
	# characters in an URL.
	# ---
	my $this   = shift;
	my $module = shift;
	my $kiwi   = $this->{kiwi};
	my $path;
	$module = $this -> quote ($module);
	$path = $this -> thisPath ($module);
	if (defined $path) {
		return $path;
	}
	$path = $this -> obsPath ($module);
	if (defined $path) {
		return $path;
	}
	$path = $this -> openSUSEpath ($module);
	if (defined $path) {
		return $path;
	}
	$path = $this -> isoPath ($module);
	if (defined $path) {
		return $path;
	}
	$path = $this -> filePath ($module);
	if (defined $path) {
		return $path;
	}
	return $module;
}

#==========================================
# quote
#------------------------------------------
sub quote {
	# ...
	# Each part of a URL, e.g. the path info,
	# the query, etc., has a different set of reserved characters that
	# must be quoted.
	#
	# RFC 2396 Uniform Resource Identifiers (URI): Generic Syntax lists
	# the following reserved characters.
	#
	# reserved    = ";" | "?" | ":" | "@" | "&" | "=" | "+" | "$" | "," | "/"
	#
	# This function will quote the user name and password of a given
	# URL string identified by .*://<user>:<pwd>@...
	# ---
	my $this = shift;
	my $surl = shift;
	my $part1;
	my $part2;
	my $part3;
	my $part4;
	$surl =~ s/^[ \t]+//g;
	$surl =~ s/[ \t]+$//g;
	if ($surl =~ /^(.*:\/\/)(.*):(.*)(\@.*)$/) {
		$part1 = $1;
		$part2 = $2;
		$part3 = $3;
		$part4 = $4;
	} else {
		return $surl;
	}
	my $safe = (
		'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_.-'
	);
	my %safe = {};
	foreach my $key (split (//,$safe)) {
		$safe{$key} = $key;
	}
	my $run = sub {
		my $part = $_[0];
		my %safe = %{$_[1]};
		my @done = ();
		foreach my $key (split (//,$part)) {
			if (! defined $safe{$key}) {
				$key = sprintf ("%%%02X",ord($key));
			}
			push @done,$key;
		}
		return join ("",@done);
	};
	$part2 = &{$run}($part2,\%safe);
	$part3 = &{$run}($part3,\%safe);
	return $part1.$part2.":".$part3.$part4;
}

#==========================================
# thisPath
#------------------------------------------
sub thisPath {
	# ...
	# This method builds a valid path from a this://..
	# description. The this path is the same as where the
	# image description tree resides
	# ---
	my $this   = shift;
	my $module = shift;
	my $kiwi;
	if (! defined $this->{kiwi}) {
		$kiwi = new KIWILog("tiny");
	} else {
		$kiwi = $this->{kiwi};
	}
	#==========================================
	# normalize URL data
	#------------------------------------------
	if ((! defined $module) || ($module !~ /^this:\/\//)) {
		return undef;
	}
	$module =~ s/this:\/\///;
	if ((! defined $module) || ($module eq "/")) {
		return undef;
	}
	my $thisPath;
	if (defined $main::ForeignRepo{prepare}) {
		if (! open FD,"$main::ForeignRepo{create}/image/main::Prepare") {
			return undef;
		}
		$thisPath = <FD>; close FD;
		$thisPath = "$thisPath/$module";
	} elsif (defined $main::Create) {
		if (! open FD,"$main::Create/image/main::Prepare") {
			return undef;
		}
		$thisPath = <FD>; close FD;
		$thisPath = "$thisPath/$module";
	} elsif (defined $main::Upgrade) {
		if (! open FD,"$main::Upgrade/image/main::Prepare") {
			return undef;
		}
		$thisPath = <FD>; close FD;
		$thisPath = "$thisPath/$module";
	} elsif (defined $main::RunTestSuite) {
		if (! open FD,"$main::RunTestSuite/image/main::Prepare") {
			return undef;
		}
		$thisPath = <FD>; close FD;
		$thisPath = "$thisPath/$module";
	} else {
		$thisPath = "$main::Prepare/$module";
	}
	if ($thisPath !~ /^\//) {
		my $pwd = qxx ("pwd"); chomp $pwd;
		$thisPath = $pwd."/".$thisPath;
	}
	return $thisPath;
}

#==========================================
# filePath
#------------------------------------------
sub filePath {
	my $this   = shift;
	my $module = shift;
	my $kiwi   = $this->{kiwi};
	#==========================================
	# normalize URL data
	#------------------------------------------
	if ((! defined $module) || ($module !~ /^file:\/\//)) {
		return undef;
	}
	$module =~ s/file:\/\///;
	if ($module !~ /^\//) {
		my $pwd = qxx ("pwd"); chomp $pwd;
		$module = $pwd."/".$module;
	}
	if (! -e $module) {
		$kiwi -> warning ("file path: $module doesn't exist: $!");
		$kiwi -> skipped ();
		return undef;
	}
	return $module;
}

#==========================================
# obsPath
#------------------------------------------
sub obsPath {
	# ...
	# This method will create an openSUSE buildservice
	# path as this:// url with the predefined subdirectories
	# images/ if called as part of a boot attribute or
	# repos if called as part of a repository source path
	# attribute
	# ---
	my $this   = shift;
	my $module = shift;
	my $boot   = shift;
	#==========================================
	# normalize URL data
	#------------------------------------------
	if ((! defined $module) || ($module !~ /^obs:\/\//)) {
		return undef;
	}
	$module =~ s/obs:\/\///;
	if ((! defined $module) || ($module eq "/")) {
		return undef;
	}
	if (defined $boot) {
		$module = "this://images/$module"
	} else {
		$module = "this://repos/$module"
	}
	my $path = $this -> thisPath ($module);
	return $path;
}

#==========================================
# isoPath
#------------------------------------------
sub isoPath {
	# ...
	# This method builds a valid path from a iso://...
	# description. The iso woll be loop mounted and
	# the loop location serves as source path
	# ---
	my $this   = shift;
	my $module = shift;
	my $kiwi   = $this->{kiwi};
	my $root   = $this->{root};
	my $result;
	my $status;
	my $file;
	my $path;
	#==========================================
	# normalize URL data
	#------------------------------------------
	if ((! defined $module) || ($module !~ /^iso:\/\//)) {
		return undef;
	}
	$module =~ s/iso:\/\///;
	#==========================================
	# Convert zypper iso URL if required
	#------------------------------------------
	if ($module =~ /iso=(.*\.iso)/) {
		$file = $1;
		if ($module =~ /url=file:\/\/(.*)/) {
			$path = $1;
		} elsif ($module =~ /url=dir:\/\/(.*)/) {
			$path = $1;
		} else {
			return undef;
		}
		$module = $path."/".$file;
	}
	#==========================================
	# Check existence of iso file
	#------------------------------------------
	if ($module !~ /^\//) {
		my $pwd = qxx ("pwd"); chomp $pwd;
		$module = $pwd."/".$module;
	}
	if (! -e $module) {
		$kiwi -> warning ("ISO path: $module doesn't exist: $!");
		$kiwi -> skipped ();
		return undef;
	}
	my $name   = basename ($module);
	my $tmpdir = "/tmp/kiwimount-$name-$$";
	#==========================================
	# create ISO mount point and perform mount
	#------------------------------------------
	if (! defined $root) {
		return $tmpdir;
	}
	$status = qxx ("mkdir -p $tmpdir 2>&1");
	$result = $? >> 8;
	if ($result != 0) {
		$kiwi -> warning ("Couldn't create tmp dir for iso mount: $status: $!");
		$kiwi -> skipped ();
		return undef;
	}
	$status = qxx ("mount -o loop $module $tmpdir 2>&1");
	$result = $? >> 8;
	if ($result != 0) {
		$kiwi -> warning ("Failed to loop mount ISO path: $status");
		$kiwi -> skipped ();
		return undef;
	}
	#==========================================
	# add loop mount to mount list of root obj
	#------------------------------------------
	$root -> addToMountList ($tmpdir);
	return $tmpdir;
}

#==========================================
# openSUSEpath
#------------------------------------------
sub openSUSEpath {
	# ...
	# This method builds a valid URL path to be used as
	# source location for an openSUSE installation source.
	# The method needs the basic openSUSE distribution or
	# module repository name information to be able to
	# complete this data into a valid path
	# ---
	my $this     = shift;
	my $module   = shift;
	my $quiet    = shift;
	my $browser  = LWP::UserAgent->new;
	my $location = $main::openSUSE;
	my @dists    = qw (standard);
	my @urllist  = ();
	my $kiwi     = $this->{kiwi};
	#==========================================
	# normalize URL data
	#------------------------------------------
	if ((! defined $module) || ($module !~ /^opensuse:\/\//)) {
		return undef;
	}
	$module =~ s/opensuse:\/\///;
	$module =~ s/\/$//;
	$module =~ s/^\///;
	if ((! defined $module) || ($module eq "")) {
		return undef;
	}
	#==========================================
	# Create urllist for later testing
	#------------------------------------------
	foreach my $dist (@dists) {
		foreach my $subdir (@main::openSUSE) {
			my $url1 = $location."/".$subdir."/".$module."/";
			push @urllist,$url1;
			if ($url1 !~ /\/$dist/) {
				my $url2 = $location."/".$subdir."/".$module."/".$dist."/";
				push @urllist,$url2;
			}
		}
	}
	#==========================================
	# Check url entries in urllist
	#------------------------------------------
	my @responses = ();
	foreach my $url (@urllist) {
		my $request = HTTP::Request->new (GET => $url);
		my $response= $browser -> request  ( $request );
		my $title   = $response-> title ();
		if ((defined $title) && ($title !~ /not found/i)) {
			my $repourl = $url;
			$request = HTTP::Request->new (GET => $repourl."/repodata");
			$response= $browser -> request  ( $request );
			$title   = $response-> title ();
			if ((defined $title) && ($title !~ /not found/i)) {
				$this->{type} = "rpm-md";
				return $url;
			} else {
				push (@responses,"$repourl/repodata -> $title");
			}
			$request = HTTP::Request->new (GET => $repourl."/media.1");
			$response= $browser -> request  ( $request );
			$title   = $response-> title ();
			if ((defined $title) && ($title !~ /not found/i)) {
				$this->{type} = "yast2";
				return $url;
			} else {
				push (@responses,"$repourl//media.1 -> $title");
			}
		} else {
			push (@responses,"$url -> $title");
		}
	}
	if (! defined $quiet) {
		my $response = join ("\n",@responses);
		$kiwi -> warning ("Couldn't resolve opensuse URL: $response");
		$kiwi -> skipped ();
	}
	return undef;
}

1;
